# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for protein secondary structure annotation using DSSP.
"""

__name__ = "biotite.application.dssp"
__author__ = "Patrick Kunzmann"

from .app import *