# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for downloading files from the RCSB PDB.
"""

__name__ = "biotite.database.rcsb"
__author__ = "Patrick Kunzmann"

from .download import *
from .query import *