# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing an :class:`AtomArray` or
:class:`AtomArrayStack` using the Gro format used by the gromacs
software package.
"""

__name__ = "biotite.structure.io.gro"
__author__ = "Daniel Bauer"

from .file import *